<?php

namespace LWVendor\PhpOffice\PhpSpreadsheet\Writer\Pdf;

use LWVendor\PhpOffice\PhpSpreadsheet\Worksheet\PageSetup;
use LWVendor\PhpOffice\PhpSpreadsheet\Writer\Pdf;
class Dompdf extends Pdf
{
    /**
     * Gets the implementation of external PDF library that should be used.
     *
     * @return \Dompdf\Dompdf implementation
     */
    protected function createExternalWriterInstance()
    {
        return new \LWVendor\Dompdf\Dompdf();
    }
    /**
     * Save Spreadsheet to file.
     *
     * @param string $pFilename Name of the file to save as
     */
    public function save($pFilename) : void
    {
        $fileHandle = parent::prepareForSave($pFilename);
        //  Default PDF paper size
        $paperSize = 'LETTER';
        //    Letter    (8.5 in. by 11 in.)
        //  Check for paper size and page orientation
        if ($this->getSheetIndex() === null) {
            $orientation = $this->spreadsheet->getSheet(0)->getPageSetup()->getOrientation() == PageSetup::ORIENTATION_LANDSCAPE ? 'L' : 'P';
            $printPaperSize = $this->spreadsheet->getSheet(0)->getPageSetup()->getPaperSize();
        } else {
            $orientation = $this->spreadsheet->getSheet($this->getSheetIndex())->getPageSetup()->getOrientation() == PageSetup::ORIENTATION_LANDSCAPE ? 'L' : 'P';
            $printPaperSize = $this->spreadsheet->getSheet($this->getSheetIndex())->getPageSetup()->getPaperSize();
        }
        $orientation = $orientation == 'L' ? 'landscape' : 'portrait';
        //  Override Page Orientation
        if ($this->getOrientation() !== null) {
            $orientation = $this->getOrientation() == PageSetup::ORIENTATION_DEFAULT ? PageSetup::ORIENTATION_PORTRAIT : $this->getOrientation();
        }
        //  Override Paper Size
        if ($this->getPaperSize() !== null) {
            $printPaperSize = $this->getPaperSize();
        }
        if (isset(self::$paperSizes[$printPaperSize])) {
            $paperSize = self::$paperSizes[$printPaperSize];
        }
        //  Create PDF
        $pdf = $this->createExternalWriterInstance();
        $pdf->setPaper(\strtolower($paperSize), $orientation);
        $pdf->loadHtml($this->generateHTMLAll());
        $pdf->render();
        //  Write to file
        \fwrite($fileHandle, $pdf->output());
        parent::restoreStateAfterSave();
    }
}
